from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Optional, TypeAlias, Union

ResourceId: TypeAlias = str


class ResourceType(Enum):
    CSV = "csv"
    JSON = "json"
    XML = "xml"
    Spreadsheet = "spreadsheet"
    NetCDF4 = "netcdf4"
    NetCDF3 = "netcdf3"
    GeoTIFF = "geotiff"
    NPDict = "np-dict"
    Shapefile = "shapefile"
    Container = "container"


@dataclass
class CSVProp:
    delimiter: str = ","


@dataclass
class Resource:
    id: ResourceId
    type: ResourceType
    prop: Optional[CSVProp] = None

    @staticmethod
    def deserialize(raw: dict):
        if raw["type"] == ResourceType.CSV.value and raw["prop"] is not None:
            prop = CSVProp(raw["prop"]["delimiter"])
        else:
            prop = None
        return Resource(raw["id"], ResourceType(raw["type"]), prop)


class ResourceData(ABC):

    @abstractmethod
    def to_dict(self):
        pass


@dataclass
class ResourceDataFile(ResourceData):
    file: str

    def to_dict(self):
        return {"file": self.file}


@dataclass
class ResourceDataString(ResourceData):
    value: Union[str, bytes]

    def to_dict(self):
        return {
            "string": (
                self.value.decode() if isinstance(self.value, bytes) else self.value
            )
        }



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, Optional

from rdflib import RDF, BNode, Graph, Literal, Namespace, URIRef

from drepr.models.drepr import DRepr
from drepr.models.sm import NodeId
from drepr.writers.base import StreamClassWriter

SubjVal = str | tuple | int | bool


class RDFGraphWriter(StreamClassWriter):
    def __init__(self, prefixes: dict[str, str]):
        self.g: Graph = Graph()
        for prefix, ns in prefixes.items():
            self.g.bind(prefix, Namespace(ns))

        self.written_records: dict[SubjVal, BNode | URIRef] = {}
        self.origin_subj: SubjVal = ""
        self.subj: Optional[URIRef | BNode] = None
        self.buffer: list[tuple[URIRef, URIRef | BNode | Literal]] = []
        self.is_buffered: bool = False
        self.has_subj_data: bool = False

    def has_written_record(self, subj: SubjVal) -> bool:
        return subj in self.written_records

    def begin_record(
        self, class_uri: str, subj: SubjVal, is_blank: bool, is_buffered: bool
    ):
        self.origin_subj = subj
        if is_blank:
            self.subj = BNode()
        else:
            # subj will be a string for URIRef
            self.subj = URIRef(subj)  # type: ignore

        if is_buffered:
            self.buffer = [(RDF.type, URIRef(class_uri))]
        else:
            self.g.add((self.subj, RDF.type, URIRef(class_uri)))
        self.is_buffered = is_buffered

    def end_record(self):
        if self.subj is None:
            # has been aborted
            return

        if len(self.buffer) > 0:
            for pred, obj in self.buffer:
                self.g.add((self.subj, pred, obj))
            self.buffer = []
        self.written_records[self.origin_subj] = self.subj
        self.subj = None

    def abort_record(self):
        """Abort the record that is being written"""
        self.subj = None
        self.buffer = []

    def is_record_empty(self):
        return not self.has_subj_data

    def write_data_property(self, predicate_id: str, value: Any, dtype: Optional[str]):
        if dtype == "drepr:uri":
            value = URIRef(value)
        else:
            value = Literal(value, datatype=dtype)

        if self.is_buffered:
            self.buffer.append((URIRef(predicate_id), value))
        else:
            assert self.subj is not None
            self.g.add((self.subj, URIRef(predicate_id), value))

    def write_object_property(
        self,
        predicate_id: str,
        object: SubjVal,
        is_subject_blank: bool,
        is_object_blank: bool,
        is_new_subj: bool,
    ):
        object = self.written_records[object]
        if self.is_buffered:
            self.buffer.append((URIRef(predicate_id), object))
        else:
            assert self.subj is not None
            self.g.add((self.subj, URIRef(predicate_id), object))

    def write_to_string(self):
        return self.g.serialize(format="ttl")

    def write_to_file(self, filepath):
        self.g.serialize(filepath, format="ttl")



from .align import *
from .attr import *
from .drepr import *
from .drepr_builder import *
from .format import OutputFormat
from .parse_v1.resource_parser import ResourceParser
from .path import *
from .preprocessing import *
from .resource import *
from .sm import *

DEFAULT_RESOURCE_ID = ResourceParser.DEFAULT_RESOURCE_ID



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from codegen.models import PredefinedFn
from codegen.models.expr import Expr, ExprFuncCall, ExprIdent, ExprVar
from codegen.models.memory import Var
from codegen.models.program import ImportManager

from drepr.models.resource import ResourceType


class DReprPredefinedFn(PredefinedFn):

    @dataclass
    class _safe_item_getter(Expr):
        collection: Expr
        item: Expr
        msg: Expr

        def to_python(self):
            return f"safe_item_getter({self.collection.to_python()}, {self.item.to_python()}, {self.msg.to_python()})"

    @dataclass
    class _safe_len(Expr):
        collection: Expr
        msg: Expr

        def to_python(self):
            return f"safe_len({self.collection.to_python()}, {self.msg.to_python()})"

    @staticmethod
    def safe_item_getter(
        import_manager: ImportManager, collection: Expr, item: Expr, msg: Expr
    ):
        import_manager.import_("drepr.utils.safe.safe_item_getter")
        return DReprPredefinedFn._safe_item_getter(collection, item, msg)

    @staticmethod
    def safe_len(import_manager: ImportManager, collection: Expr, msg: Expr):
        import_manager.import_("drepr.utils.safe.safe_len")
        return DReprPredefinedFn._safe_len(collection, msg)

    @staticmethod
    def read_source(
        import_manager: ImportManager, source_type: ResourceType, input_file: Var
    ):
        import_manager.import_(f"drepr.readers.prelude.read_source_{source_type.value}")
        return ExprFuncCall(
            ExprIdent(f"read_source_{source_type.value}"), [ExprVar(input_file)]
        )



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional, TypeAlias, Union

from .attr import Attr, AttrId

DREPR_URI = "https://purl.org/drepr/1.0/uri"
NodeId: TypeAlias = str
EdgeId: TypeAlias = int


class DataType(Enum):
    xsd_decimal = "xsd:decimal"
    xsd_anyURI = "xsd:anyURI"
    xsd_gYear = "xsd:gYear"
    xsd_date = "xsd:date"
    xsd_dateTime = "xsd:dateTime"
    xsd_int = "xsd:int"
    xsd_string = "xsd:string"
    geo_wktLiteral = "geo:wktLiteral"
    drepr_uri = "drepr:uri"


@dataclass
class ClassNode:
    node_id: NodeId
    label: str  # relative iri

    def get_abs_iri(self, sm: SemanticModel):
        """Get the absolute IRI of this node"""
        if sm.is_rel_iri(self.label):
            return sm.get_abs_iri(self.label)
        return self.label

    def get_rel_iri(self, sm: SemanticModel):
        if sm.is_rel_iri(self.label):
            return self.label
        return sm.get_rel_iri(self.label)

    def is_blank_node(self, sm: SemanticModel) -> bool:
        for e in sm.iter_outgoing_edges(self.node_id):
            if e.get_abs_iri(sm) == DREPR_URI:
                return False
        return True


@dataclass
class DataNode:
    node_id: NodeId
    attr_id: AttrId
    data_type: Optional[DataType] = None


@dataclass
class LiteralNode:
    node_id: NodeId
    # you should rely on data_type to get the type of value right. The parser may be wrong about it.
    value: Any
    data_type: Optional[DataType] = None
    # whether to always generate values of the literal node, even if all the other non-literal nodes are missing
    # however, if the parent class node has URI and the URI is missing, we won't generate the literal node
    always_generate: bool = False


@dataclass
class Edge:
    edge_id: EdgeId
    source_id: NodeId
    target_id: NodeId
    label: str  # rel uri
    is_subject: bool = False
    is_required: bool = False

    def get_abs_iri(self, sm: SemanticModel):
        """Get the absolute IRI of the predicate"""
        if sm.is_rel_iri(self.label):
            return sm.get_abs_iri(self.label)
        return self.label

    def get_rel_iri(self, sm: SemanticModel):
        if sm.is_rel_iri(self.label):
            return self.label
        return sm.get_rel_iri(self.label)


Node = Union[LiteralNode, DataNode, ClassNode]


@dataclass
class SemanticModel:
    nodes: dict[NodeId, Node]
    edges: dict[EdgeId, Edge]
    prefixes: dict[str, str]

    @staticmethod
    def get_default(attrs: list[Attr]) -> SemanticModel:
        """
        Automatically generate a semantic model from a list of attributes.

        WARNING: the engine may not able to map data to this semantic model if the final output should be
        comprised of multiple tables.
        """
        prefixes = {"eg": "https://example.org/"}
        aids = {attr.id for attr in attrs}
        cid = None
        for i in range(len(attrs)):
            cid = f"c{i}"
            if cid not in aids:
                break
        assert cid is not None
        nodes: dict[str, Node] = {cid: ClassNode(cid, "eg:Record")}
        edges = {}
        for attr in attrs:
            nodes[attr.id] = DataNode(attr.id, attr.id, None)
            edge_id = len(edges)
            edges[edge_id] = Edge(edge_id, cid, attr.id, f"eg:{attr.id}")

        return SemanticModel(nodes, edges, prefixes)

    @staticmethod
    def get_default_prefixes() -> dict[str, str]:
        return {
            "drepr": "https://purl.org/drepr/1.0/",
            "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
            "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
            "owl": "http://www.w3.org/2002/07/owl#",
        }

    @staticmethod
    def deserialize(raw: dict) -> SemanticModel:
        nodes = {}
        for nid, n in raw["nodes"].items():
            if n["type"] == "class_node":
                nodes[nid] = ClassNode(n["node_id"], n["label"])
            elif n["type"] == "data_node":
                nodes[nid] = DataNode(
                    n["node_id"],
                    n["attr_id"],
                    DataType(n["data_type"]) if n["data_type"] is not None else None,
                )
            elif n["type"] == "literal_node":
                nodes[nid] = LiteralNode(
                    n["node_id"],
                    n["value"],
                    DataType(n["data_type"]) if n["data_type"] is not None else None,
                )
            else:
                raise NotImplementedError()
        edges = {eid: Edge(**e) for eid, e in raw["edges"].items()}
        return SemanticModel(nodes, edges, raw["prefixes"])

    def get_class_node(self, node_id: str) -> ClassNode:
        node = self.nodes[node_id]
        if not isinstance(node, ClassNode):
            raise ValueError(f"The node {node_id} is not a class node")
        return node

    def remove_node(self, node_id: str):
        self.nodes.pop(node_id)
        removed_edges = []
        for eid, e in self.edges.items():
            if e.source_id == node_id or e.target_id == node_id:
                removed_edges.append(eid)
        for eid in removed_edges:
            self.edges.pop(eid)

    def class2dict(self, class_id: str) -> dict[str, Union[list[int], int]]:
        """
        Get a dictionary that contains information (predicates) about a given class
        """
        info = {}
        for eid, e in self.edges.items():
            if e.source_id != class_id:
                continue

            if e.label in info:
                if not isinstance(info[e.label], list):
                    info[e.label] = [info[e.label], eid]
                else:
                    info[e.label].append(eid)
            else:
                info[e.label] = eid
        return info

    def iter_class_nodes(self):
        for n in self.nodes.values():
            if isinstance(n, ClassNode):
                yield n

    def iter_outgoing_edges(self, node_id: str):
        for e in self.edges.values():
            if e.source_id == node_id:
                yield e

    def iter_incoming_edges(self, node_id: str):
        for e in self.edges.values():
            if e.target_id == node_id:
                yield e

    def iter_child_nodes(self, node_id: str):
        for e in self.edges.values():
            if e.source_id == node_id:
                yield self.nodes[e.source_id]

    def iter_parent_nodes(self, node_id: str):
        for e in self.edges.values():
            if e.target_id == node_id:
                yield self.nodes[e.target_id]

    def get_rel_iri(self, abs_iri: str) -> str:
        """Convert an absolute IRI to a relative IRI."""
        assert not self.is_rel_iri(abs_iri)
        for prefix, uri in self.prefixes.items():
            if abs_iri.startswith(uri):
                return f"{prefix}:{abs_iri.replace(uri, '')}"
        raise ValueError(
            "Cannot create relative IRI because there is no suitable prefix"
        )

    def get_abs_iri(self, rel_iri: str) -> str:
        """Convert a relative IRI to an absolute IRI."""
        assert self.is_rel_iri(rel_iri)
        prefix, val = rel_iri.split(":", 1)
        if prefix not in self.prefixes:
            raise ValueError(
                f"Cannot create absolute IRI because the prefix {prefix} does not exist"
            )
        return f"{self.prefixes[prefix]}{val}"

    def is_rel_iri(self, iri: str) -> bool:
        return iri.find("://") == -1 and iri.find(":") != -1

    def get_n_class_nodes(self) -> int:
        return sum(1 for _ in self.iter_class_nodes())

    def get_edge_between_nodes(self, source_id: str, target_id: str) -> Optional[Edge]:
        matched_edges = []
        for e in self.edges.values():
            if e.source_id == source_id and e.target_id == target_id:
                matched_edges.append(e)

        if len(matched_edges) == 0:
            return None
        elif len(matched_edges) == 1:
            return matched_edges[0]
        else:
            raise ValueError(
                f"Found multiple edges between {source_id} and {target_id}"
            )

    def get_edges_between_nodes(self, source_id: str, target_id: str):
        matched_edges = []
        for e in self.edges.values():
            if e.source_id == source_id and e.target_id == target_id:
                matched_edges.append(e)
        return matched_edges



import re
from abc import ABC, abstractmethod
from typing import List, Union

from drepr.utils.validator import InputError
from ..path import Path, IndexExpr, RangeExpr, WildcardExpr, Expr
from ..resource import Resource


class PathParser(ABC):
    @abstractmethod
    def parse(self, resource: Resource, path: Union[str, list], parse_trace: str) -> Path:
        pass

    # noinspection PyMethodMayBeStatic
    def get_resource(self, resources: List[Resource], resource_id: str, trace: str) -> Resource:
        for res in resources:
            if res.id == resource_id:
                return res
        raise InputError(f"{trace}\nERROR: Refer to path of an nonexistent resource: {resource_id}")


class PathParserV1(PathParser):
    """
    A path can either be a JSONPath or our list path

    1. If the path is a JSONPath, then it is a string startswith `$`. We only support the following
       type of step: range, index, list of index, and wildcard. However, wildcard is only used for selecting
       all values of an object
    2. If the path is a normal string
    """
    REG_SRANGE = re.compile(r"^(\d+)?\.\.(-?\d+)?(?::(\d+))?$")
    REG_SINDEX = re.compile(r"^(?:\$\{([^}]+)})|(\d+)|(.*)$")
    REG_SRANGE_EXPR = re.compile(
        r"^(?:(\d+)|(?:\$\{([^}]+)}))?\.\.(?:(-\d+)|(?:\$\{([^}]+)}))?(?::(\d+)|(?:\$\{([^}]+)}))?$"
    )

    REG_JPATH_BRACKET = re.compile(r"(?:\[(-?\d+)?\:(?:(-?\d+)(?:\:(-?\d+))?)?\])|(?:\[(-?\d+)\])|(?:\['([^']+)'\])")
    REG_JPATH_DOT = re.compile(r"\.((?:(?!\.|\[).)+)")

    def parse(self, _resource: Resource, path: Union[str, list], parse_trace: str) -> Path:
        if isinstance(path, str):
            return self.parse_jsonpath(path, parse_trace)

        if isinstance(path, list):
            return self.parse_custom_path(path, parse_trace)

        raise InputError(f"{parse_trace}\nERROR: the path must either be a "
                         f"string (JSONPath) or a list of steps. Get {type(path)} instead")

    def parse_jsonpath(self, jpath: str, parse_trace: str) -> Path:
        if not jpath.startswith("$"):
            raise InputError(f"{parse_trace}\nERROR: invalid json path. The path must start with `$`. "
                             f"Get: {jpath}")

        jpath = jpath[1:]
        steps = []
        parsing_pos = 1

        while len(jpath) > 0:
            if jpath.startswith("["):
                m = self.REG_JPATH_BRACKET.match(jpath)
                if m is None:
                    raise InputError(
                        f"{parse_trace}\nERROR: invalid json path, error while parsing bracket at position {parsing_pos}")

                jpath = jpath[m.span()[-1]:]
                parsing_pos += m.span()[-1]  # m.span()[0] is always 0

                if m.group(5) is not None:
                    # match with string
                    steps.append(IndexExpr(m.group(5)))
                elif m.group(4) is not None:
                    # match with a single number
                    steps.append(IndexExpr(int(m.group(4))))
                else:
                    steps.append(RangeExpr(
                        int(m.group(1) or "0"),
                        int(m.group(2)) if m.group(2) is not None else None,
                        int(m.group(3) or "1")))
            elif jpath.startswith(".*~"):
                # *~ select property names
                steps.append(WildcardExpr.Names)
                jpath = jpath[3:]
                parsing_pos += 3
            elif jpath.startswith(".*"):
                steps.append(WildcardExpr.Values)
                jpath = jpath[2:]
                parsing_pos += 2
            else:
                m = self.REG_JPATH_DOT.match(jpath)
                if m is None:
                    raise InputError(
                        f"{parse_trace}\nERROR: invalid json path, error while parsing step at position {parsing_pos}")

                jpath = jpath[m.span()[-1]:]
                parsing_pos += m.span()[-1]  # m.span()[0] is always 0

                # after a dot, it can either be a number or a string
                if m.group(1).isdigit():
                    steps.append(IndexExpr(int(m.group(1))))
                else:
                    steps.append(IndexExpr(m.group(1)))

        return Path(steps)

    def parse_custom_path(self, path: List[str], parse_trace: str) -> Path:
        steps = []
        for i, step in enumerate(path):
            trace = f"Parsing step {i} ({step})"
            if isinstance(step, str):
                m = self.REG_SRANGE.match(step)
                if m is not None:
                    steps.append(RangeExpr(
                        int(m.group(1) or '0'),
                        int(m.group(2)) if m.group(2) is not None else None,
                        int(m.group(3) or '1')))
                    continue

                m = self.REG_SRANGE_EXPR.match(step)
                if m is not None:
                    steps.append(RangeExpr(
                        (Expr(m.group(1)[2:-1]) if m.group(1).startswith("${") else int(m.group(1)))
                        if m.group(1) is not None else 0,
                        (Expr(m.group(2)[2:-1]) if m.group(2).startswith("${") else int(m.group(2)))
                        if m.group(2) is not None else None,
                        (Expr(m.group(2)[2:-1]) if m.group(2).startswith("${") else int(m.group(2)))
                        if m.group(2) is not None else 1))
                    continue

                if step.startswith("${"):
                    steps.append(IndexExpr(Expr(step[2:-1])))
                else:
                    steps.append(IndexExpr(step))
            elif isinstance(step, int):
                steps.append(IndexExpr(step))
            else:
                raise InputError(
                    f"{parse_trace}\n{trace}\nERROR: step must either be string or number. Get {type(step)} instead")

        return Path(steps)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from functools import partial
from pathlib import Path
from typing import Callable

from codegen.models import AST, Memory, PredefinedFn, Program, Var, VarScope, expr, stmt
from codegen.models.program import ImportManager
from typing_extensions import is_protocol

from drepr.models.path import IndexExpr
from drepr.models.prelude import Cardinality, DRepr, OutputFormat
from drepr.models.preprocessing import PMap, Preprocessing, PreprocessingType
from drepr.planning.class_map_plan import (
    BlankObject,
    BlankSubject,
    ClassesMapExecutionPlan,
    ClassMapPlan,
    DataProp,
    ExternalIDSubject,
    IDObject,
    InternalIDSubject,
    ObjectProp,
    Subject,
)
from drepr.program_generation.alignment_fn import AlignmentFn, PathAccessor
from drepr.program_generation.predefined_fn import DReprPredefinedFn
from drepr.program_generation.program_space import VarSpace
from drepr.program_generation.writers import Writer
from drepr.utils.misc import assert_not_null, assert_true


@dataclass
class FileOutput:
    fpath: Path
    format: OutputFormat


@dataclass
class MemoryOutput:
    format: OutputFormat


Output = FileOutput | MemoryOutput


def gen_program(
    desc: DRepr, exec_plan: ClassesMapExecutionPlan, output: Output, debuginfo: bool
) -> AST:
    """Generate a program to convert the given D-REPR to a target format"""
    program = Program()
    mem = program.memory
    writer = Writer(desc, output.format)

    func_args = [
        Var.create(mem, "resource", key=VarSpace.resource(res.id))
        for res in desc.resources
    ]
    if isinstance(output, FileOutput):
        output_file = Var.create(mem, "output_file", key=VarSpace.output_file())
        func_args.append(output_file)
    else:
        output_file = None

    program.root.linebreak()
    main_fn = program.root.func("main", func_args)

    for resource in desc.resources:
        var = Var.create(mem, "resource_data", key=VarSpace.resource_data(resource.id))
        main_fn.assign(
            mem,
            var,
            DReprPredefinedFn.read_source(
                program.import_manager,
                resource.type,
                Var.deref(mem, key=VarSpace.resource(resource.id)),
            ),
        )

    # define missing values of attributes
    main_fn.linebreak()
    for attr in desc.attrs:
        if len(attr.missing_values) > 0:
            main_fn.assign(
                mem,
                Var.create(
                    mem,
                    f"{attr.id}_missing_values",
                    key=VarSpace.attr_missing_values(attr.id),
                ),
                expr.ExprConstant(attr.missing_values),
            )

    # create transformation
    for preprocess in desc.preprocessing:
        gen_preprocess_executor(program, desc, main_fn, preprocess)

    # create a writer
    writer.create_writer(program.import_manager, mem, main_fn)

    # for each class node, we generate a plan for each of them.
    for classplan in exec_plan.class_map_plans:
        main_fn.linebreak()
        main_fn.comment(f"Transform records of class {classplan.class_id}")

        varscope = main_fn.next_var_scope()

        # generate the code to execute the plan
        gen_classplan_executor(
            program.import_manager, mem, main_fn, writer, desc, classplan, debuginfo
        )

        # claim the variables that aren't going to be used anymore
        for var in Var.find_by_scope(mem, varscope):
            var.delete(mem)

    main_fn.linebreak()
    # we write the output to the file
    if isinstance(output, FileOutput):
        assert output_file is not None
        writer.write_to_file(mem, main_fn, expr.ExprVar(output_file))
    else:
        content = Var.create(mem, "output", key=tuple())
        writer.write_to_string(mem, main_fn, content)
        main_fn.return_(expr.ExprVar(content))

    invok_main = expr.ExprFuncCall(
        expr.ExprIdent("main"), [expr.ExprIdent("*sys.argv[1:]")]
    )

    program.root.linebreak()
    program.root.if_(
        expr.ExprEqual(expr.ExprIdent("__name__"), expr.ExprConstant("__main__"))
    )(
        stmt.ImportStatement("sys"),
        stmt.LineBreak(),
        stmt.SingleExprStatement(
            expr.ExprFuncCall(expr.ExprIdent("print"), [invok_main])
            if isinstance(output, MemoryOutput)
            else invok_main
        ),
    )
    return program.root


def gen_classplan_executor(
    import_manager: ImportManager,
    mem: Memory,
    parent_ast: AST,
    writer: Writer,
    desc: DRepr,
    classplan: ClassMapPlan,
    debuginfo: bool,
):
    """Generate the code to execute the given class plan.
    Below is the pseudo code:

    1. Iterate over the subject values
        1. If the subject is uri and it has missing values, if the uri is missing, we skip this
           record
        2. Begin record
        3. Iterate over target property & value
            1. If not target.can_have_missing_values:
                1. Iterate over objprop values:
                    1. Write property
            2. Else:
                1. If target edge is optional:
                    iterate over objprop values:
                        if objprop value is not missing:
                            write  property
                else:
                    (1) ----
                    has_record = False
                    iterate over objprop values:
                        if objprop value is not missing:
                            has_record = True
                            write property
                    if not has_record:
                        abort the record
                    ---- (2)
        4. End the record -- if the subject is blank node,
            and we do not write any data, we abort, otherwise, we commit
    """
    class_uri = expr.ExprConstant(
        desc.sm.get_abs_iri(desc.sm.get_class_node(classplan.class_id).label)
    )

    def on_missing_key(tree: AST):
        if parent_ast.has_statement_between_ast(stmt.ForLoopStatement, tree.id):
            tree(stmt.ContinueStatement())
        else:
            # same ast because of a single value, we can't use continue
            # however, we use pass as it's a single-level if/else -- the else part
            # will handle the instance generation if there is no missing value.
            tree(stmt.NoStatement())

    ast = PathAccessor(import_manager).iterate_elements(
        mem,
        parent_ast,
        classplan.subject.attr,
        None,
        None,
        validate_path=debuginfo,
        on_missing_key=(
            on_missing_key if classplan.subject.attr.path.has_optional_steps() else None
        ),
    )
    is_subj_blank = isinstance(classplan.subject, BlankSubject)
    can_class_missing = (
        any(
            not dprop.is_optional and dprop.can_target_missing
            for dprop in classplan.data_props
        )
        or any(
            not oprop.is_optional and not oprop.can_target_missing
            for oprop in classplan.object_props
        )
        or any(
            not oprop.is_optional and not oprop.can_target_missing
            for oprop in classplan.buffered_object_props
        )
    )
    is_buffered = can_class_missing

    if isinstance(classplan.subject, (InternalIDSubject, ExternalIDSubject)):
        get_subj_val = lambda: expr.ExprVar(
            Var.deref(
                mem,
                key=VarSpace.attr_value_dim(
                    classplan.subject.attr.resource_id,
                    classplan.subject.attr.id,
                    len(classplan.subject.attr.path.steps) - 1,
                ),
            )
        )
    else:
        # if this is a blank node, the subj_val is the entire index that leads to the last value
        _non_index_steps = [
            expr.ExprConstant(desc.get_attr_index_by_id(classplan.subject.attr.id))
        ] + [
            expr.ExprVar(
                Var.deref(
                    mem,
                    key=VarSpace.attr_index_dim(
                        classplan.subject.attr.resource_id,
                        classplan.subject.attr.id,
                        dim,
                    ),
                )
            )
            for dim, step in enumerate(classplan.subject.attr.path.steps)
            if not isinstance(step, IndexExpr)
        ]
        get_subj_val = lambda: (PredefinedFn.tuple(_non_index_steps))

    if (
        isinstance(classplan.subject, (InternalIDSubject, ExternalIDSubject))
        and len(classplan.subject.attr.missing_values) > 0
    ):
        # we know immediately that it's missing if the URI is missing

        if ast.id == parent_ast.id:
            # same ast because of a single value, we can't use continue
            # so we wrap it with if -- if not missing, continue to generate the instance
            ast = ast.if_(
                expr.ExprNegation(
                    PredefinedFn.set_contains(
                        expr.ExprVar(
                            Var.deref(
                                mem,
                                key=VarSpace.attr_missing_values(
                                    classplan.subject.attr.id
                                ),
                            )
                        ),
                        get_subj_val(),
                    )
                )
            )
        else:
            ast.if_(
                PredefinedFn.set_contains(
                    expr.ExprVar(
                        Var.deref(
                            mem,
                            key=VarSpace.attr_missing_values(classplan.subject.attr.id),
                        )
                    ),
                    get_subj_val(),
                )
            )(stmt.ContinueStatement())

    writer.begin_record(
        mem,
        ast,
        class_uri,
        get_subj_val(),
        expr.ExprConstant(is_subj_blank),
        is_buffered,
    )

    for dataprop in classplan.data_props:
        ast.linebreak()
        ast.comment(f"Retrieve value of data property: {dataprop.attr.id}")

        gen_classprop_body(
            import_manager,
            desc,
            mem,
            ast,
            writer,
            is_buffered,
            is_subj_blank,
            dataprop,
            debuginfo,
        )

    for objprop in classplan.object_props:
        ast.linebreak()
        ast.comment(f"Retrieve value of object property: {objprop.attr.id}")

        gen_classprop_body(
            import_manager,
            desc,
            mem,
            ast,
            writer,
            is_buffered,
            is_subj_blank,
            objprop,
            debuginfo,
        )

    assert len(classplan.buffered_object_props) == 0, "Not implemented yet"

    # we can end the record even if we abort it before. the end record code should handle this.
    ast.linebreak()

    if isinstance(classplan.subject, BlankSubject) and can_class_missing:
        ast.if_(writer.is_record_empty(mem, ast))(
            lambda ast00: writer.abort_record(mem, ast00)
        )
    else:
        writer.end_record(mem, ast)

    return ast


def gen_classprop_body(
    import_manager: ImportManager,
    desc: DRepr,
    mem: Memory,
    ast: AST,
    writer: Writer,
    is_buffered: bool,
    is_subj_blank: bool,
    classprop: DataProp | ObjectProp,
    debuginfo: bool,
):
    attr = classprop.attr
    iter_final_list = False
    if isinstance(classprop, (DataProp, IDObject)):
        if isinstance(classprop, DataProp) and classprop.attr.value_type.is_list():
            # for a list, we need to iterate over the list.
            get_prop_val = lambda: expr.ExprVar(
                Var.deref(
                    mem,
                    key=VarSpace.attr_value_dim(
                        attr.resource_id,
                        attr.id,
                        len(
                            attr.path.steps
                        ),  # not -1 because the last dimension is now a list
                    ),
                )
            )
            iter_final_list = True
        else:
            get_prop_val = lambda: expr.ExprVar(
                Var.deref(
                    mem,
                    key=VarSpace.attr_value_dim(
                        attr.resource_id,
                        attr.id,
                        len(attr.path.steps) - 1,
                    ),
                )
            )
    else:
        assert isinstance(classprop, BlankObject)
        get_prop_val = lambda: (
            PredefinedFn.tuple(
                [expr.ExprConstant(desc.get_attr_index_by_id(classprop.attr.id))]
                + [
                    expr.ExprVar(
                        Var.deref(
                            mem,
                            key=VarSpace.attr_index_dim(
                                classprop.attr.resource_id,
                                classprop.attr.id,
                                dim,
                            ),
                        )
                    )
                    for dim, step in enumerate(classprop.attr.path.steps)
                    if not isinstance(step, IndexExpr)
                ]
            )
        )

    if isinstance(classprop, DataProp):
        if len(attr.missing_values) == 0:
            # leverage the fact that if True will be optimized away
            is_prop_val_not_missing = lambda: expr.ExprConstant(True)
        else:
            is_prop_val_not_missing = lambda: PredefinedFn.set_contains(
                expr.ExprNegation(
                    expr.ExprVar(
                        Var.deref(
                            mem,
                            key=VarSpace.attr_missing_values(attr.id),
                        )
                    )
                ),
                get_prop_val(),
            )
        write_fn = partial(
            writer.write_data_property, dtype=expr.ExprConstant(classprop.datatype)
        )
    else:
        assert isinstance(classprop, ObjectProp)
        is_prop_val_not_missing = lambda: writer.has_written_record(
            mem,
            get_prop_val(),
        )
        write_fn = partial(
            writer.write_object_property,
            is_subject_blank=expr.ExprConstant(is_subj_blank),
            is_object_blank=expr.ExprConstant(isinstance(classprop, BlankObject)),
            is_new_subj=expr.ExprConstant(False),
        )

    if not classprop.can_target_missing:
        AlignmentFn(desc, import_manager).align(
            mem, ast, classprop.alignments, debuginfo, None, iter_final_list
        )(
            lambda ast_l0: write_fn(
                mem,
                ast_l0,
                expr.ExprConstant(classprop.predicate),
                get_prop_val(),
            )
        )
    else:
        if classprop.is_optional:
            AlignmentFn(desc, import_manager).align(
                mem,
                ast,
                classprop.alignments,
                debuginfo,
                # if the value is missing, we just ignore it.
                on_missing_key=lambda astxx: astxx(stmt.NoStatement()),
                iter_final_list=iter_final_list,
            )(
                lambda ast00: ast00.if_(is_prop_val_not_missing())(
                    lambda ast01: write_fn(
                        mem,
                        ast01,
                        expr.ExprConstant(classprop.predicate),
                        get_prop_val(),
                    )
                )
            )
        else:
            if classprop.alignments_cardinality.is_star_to_many():
                has_dataprop_val = Var.create(
                    mem,
                    f"{attr.id}_has_value_d{len(attr.path.steps) - 1}",
                    key=VarSpace.has_attr_value_dim(
                        attr.resource_id,
                        attr.id,
                        len(attr.path.steps) - 1,
                    ),
                )
                ast.assign(mem, has_dataprop_val, expr.ExprConstant(False))
                AlignmentFn(desc, import_manager).align(
                    mem,
                    ast,
                    classprop.alignments,
                    debuginfo,
                    lambda astxx: astxx(stmt.NoStatement()),
                    iter_final_list,
                )(
                    lambda ast00: ast00.if_(is_prop_val_not_missing())(
                        lambda ast01: ast01.assign(
                            mem, has_dataprop_val, expr.ExprConstant(True)
                        ),
                        lambda ast02: write_fn(
                            mem,
                            ast02,
                            expr.ExprConstant(classprop.predicate),
                            get_prop_val(),
                        ),
                    )
                )
                ast.if_(expr.ExprNegation(expr.ExprVar(has_dataprop_val)))(
                    lambda ast00: (
                        assert_true(
                            is_buffered,
                            "We should only abort record if we are buffering",
                        )
                        and writer.abort_record(mem, ast00)
                    )
                )
            else:
                AlignmentFn(desc, import_manager).align(
                    mem,
                    ast,
                    classprop.alignments,
                    debuginfo,
                    on_missing_key=lambda astxx: assert_true(
                        is_buffered,
                        "We should only abort record if we are buffering",
                    )
                    and writer.abort_record(mem, astxx),
                    iter_final_list=iter_final_list,
                )(
                    lambda ast00: ast00.if_(is_prop_val_not_missing())(
                        lambda ast01: write_fn(
                            mem,
                            ast01,
                            expr.ExprConstant(classprop.predicate),
                            get_prop_val(),
                        ),
                    ),
                    lambda ast10: ast10.else_()(
                        lambda ast11: (
                            assert_true(
                                is_buffered,
                                "We should only abort record if we are buffering",
                            )
                            and writer.abort_record(mem, ast11)
                        ),
                    ),
                )


def gen_preprocess_executor(
    program: Program, desc: DRepr, ast: AST, preprocessing: Preprocessing
) -> None:
    if preprocessing.type == PreprocessingType.pmap:
        value = preprocessing.value
        assert isinstance(value, PMap)

        # we don't support change_structure and output yet
        assert not value.change_structure and value.output is None

        raise NotImplementedError()

        program.root.linebreak()

        # define the transformation functionoc

        return

    raise NotImplementedError(preprocessing.type)
